"""Simple module doing its best as taking a screenshot of the current screen"""

import os
import sys
import time

from contextlib import ExitStack
from datetime import datetime
from functools import partial
from subprocess import CalledProcessError, DEVNULL, check_call


def take_screenshot(output_file=None):
    """
    Simple function trying to take a screenshot using various famous back-end programs.
    When supported by the found and available back-end, try to honor `output_file`.
    """
    if not output_file or os.path.isdir(output_file):
        # When a directory is provided, we've to force `output_file` to represent a **file** path.
        output_file = os.path.join(
            (output_file or os.getcwd()),
            datetime.now().strftime('archey4_screenshot_%Y-%m-%d_%H.%M.%S.png')
        )

    # Some programs don't accept specific filename as parameters.
    # In such cases, we may provide them a target directory instead.
    output_dir = os.path.dirname(output_file)

    # Back-end programs that _may_ (?) be available across different platforms.
    screenshot_tools = {
        'Flameshot': ['flameshot', 'full', '-p', output_dir],
        'ImageMagick': ['import', '-window', 'root', output_file],
        'scrot': ['scrot', '-z', output_file],
        'Shutter': ['shutter', '-f', '-o', output_file, '-e'],
    }

    # Extends the original screenshot tools dictionary according to current platform.
    if sys.platform in ('win32', 'cygwin'):
        screenshot_tools['SnippingTool'] = ['SnippingTool.exe', '/clip']
    elif sys.platform == 'darwin':
        screenshot_tools['ScreenCapture'] = [
            'screencapture',
            '-x',
            '-t', output_file.rpartition('.')[2],
            output_file
        ]
    else:  # *NIX systems (and others)...
        screenshot_tools['Escrotum'] = ['escrotum', output_file]
        screenshot_tools['GNOME-Screenshot'] = ['gnome-screenshot', '-f', output_file]
        screenshot_tools['grim'] = ['grim', output_file]
        screenshot_tools['KDE-Spectacle'] = ['spectacle', '-b', '-o', output_file]
        screenshot_tools['Xfce4-Screenshooter'] = ['xfce4-screenshooter', '-f', '-s', output_dir]
        screenshot_tools['Screencap (Android)'] = [
            'screencap',  # Binary available on Android.
            '-p',         # It only accepts PNG as image output format.
            (output_file.rpartition('.')[0] + '.png')
        ]

    # This part purposefully blocks so we wait a little bit before taking the screenshot.
    # It prevents taking a screenshot before Archey's output has appeared.
    for time_remaining in range(3, 0, -1):
        taking_sc_str = 'Taking screenshot in {:1d}...'.format(time_remaining)
        print(taking_sc_str, end='', flush=True)
        time.sleep(1)
        print('\r' + ' ' * len(taking_sc_str), end='\r', flush=True)
    time.sleep(0.5)

    with ExitStack() as defer_stack:
        for screenshot_tool, screenshot_cmd in screenshot_tools.items():
            try:
                check_call(screenshot_cmd, stderr=DEVNULL)
            except FileNotFoundError:
                continue
            except CalledProcessError as process_error:
                defer_stack.callback(partial(
                    print,
                    'Couldn\'t take a screenshot with {}: \"{}\".'.format(
                        screenshot_tool, process_error
                    ),
                    file=sys.stderr
                ))
                continue
            break
        else:
            defer_stack.callback(partial(
                sys.exit,
                """\
Sorry, we couldn\'t find any supported program to take a screenshot on your system.
Please install one of the following and try again: {}.\
""".format(', '.join(screenshot_tools.keys()))))



"""Test module for Archey's device's model detection module"""

from subprocess import CalledProcessError

import unittest
from unittest.mock import MagicMock, mock_open, patch

from archey.entries.model import Model
from archey.test.entries import HelperMethods
from archey.constants import DEFAULT_CONFIG


class TestModelEntry(unittest.TestCase):
    """
    For this test we have to go through several eventualities :
    * Laptop / Desktop "regular" environments
    * Raspberry Pi
    * Virtual environment (as a VM or a container)
    * Android devices
    """
    @patch('archey.entries.model.os.getuid')
    @patch('archey.entries.model.check_output')
    @HelperMethods.patch_clean_configuration
    def test_fetch_virtual_env_info(self, check_output_mock, getuid_mock):
        """Test `_fetch_virtual_env_info` method"""
        model_mock = HelperMethods.entry_mock(Model)

        with self.subTest('Detected virtual environment.'):
            check_output_mock.side_effect = [
                FileNotFoundError(),  # `systemd-detect-virt` is not available.
                'xen\nxen-domU\n',    # `virt-what` example output.
                'HYPERVISOR-NAME\n'   # `dmidecode` example output.
            ]
            getuid_mock.return_value = 0

            self.assertEqual(
                Model._fetch_virtual_env_info(model_mock),  # pylint: disable=protected-access
                'HYPERVISOR-NAME (xen, xen-domU)'
            )

        with self.subTest('Virtual environment without `dmidecode`.'):
            check_output_mock.reset_mock()
            getuid_mock.reset_mock()
            check_output_mock.side_effect = [
                FileNotFoundError(),  # `systemd-detect-virt` is not available.
                'xen\nxen-domU\n',    # `virt-what` example output.
                FileNotFoundError()   # `dmidecode` will fail.
            ]
            getuid_mock.return_value = 0

            self.assertEqual(
                Model._fetch_virtual_env_info(model_mock),  # pylint: disable=protected-access
                DEFAULT_CONFIG['default_strings']['virtual_environment'] + ' (xen, xen-domU)'
            )

        with self.subTest('Virtual environment with systemd only.'):
            check_output_mock.reset_mock()
            getuid_mock.reset_mock()
            check_output_mock.side_effect = [
                'systemd-nspawn\n'  # `systemd-detect-virt` output.
            ]
            getuid_mock.return_value = 1000  # `virt-what` and `dmidecode` won't be called.

            self.assertEqual(
                Model._fetch_virtual_env_info(model_mock),  # pylint: disable=protected-access
                DEFAULT_CONFIG['default_strings']['virtual_environment'] + ' (systemd-nspawn)'
            )

        with self.subTest('Virtual environment with systemd and `dmidecode`.'):
            check_output_mock.reset_mock()
            getuid_mock.reset_mock()
            check_output_mock.side_effect = [
                'systemd-nspawn\n',  # `systemd-detect-virt` example output.
                                     # `virt-what` won't be called (systemd call succeeded).
                'HYPERVISOR-NAME\n'  # `dmidecode` example output.
            ]
            getuid_mock.return_value = 0

            self.assertEqual(
                Model._fetch_virtual_env_info(model_mock),  # pylint: disable=protected-access
                'HYPERVISOR-NAME (systemd-nspawn)'
            )

        with self.subTest('Not a virtual environment (systemd).'):
            check_output_mock.reset_mock()
            getuid_mock.reset_mock()
            check_output_mock.side_effect = CalledProcessError(1, 'systemd-detect-virt', 'none\n')

            self.assertIsNone(
                Model._fetch_virtual_env_info(model_mock)  # pylint: disable=protected-access
            )

        with self.subTest('Not a virtual environment (virt-what).'):
            check_output_mock.reset_mock()
            getuid_mock.reset_mock()
            check_output_mock.side_effect = [
                FileNotFoundError(),  # `systemd-detect-virt` won't be available.
                '\n'                  # `virt-what` won't detect anything.
                                      # `dmidecode` won't even be called.
            ]
            getuid_mock.return_value = 0

            self.assertIsNone(
                Model._fetch_virtual_env_info(model_mock)  # pylint: disable=protected-access
            )

        with self.subTest('Not a virtual environment (no tools, no root)'):
            check_output_mock.reset_mock()
            getuid_mock.reset_mock()
            check_output_mock.side_effect = [
                FileNotFoundError()  # `systemd-detect-virt` won't be available.
            ]
            getuid_mock.return_value = 1000  # `virt-what` and `dmidecode` won't be called.

            self.assertIsNone(
                Model._fetch_virtual_env_info(model_mock)  # pylint: disable=protected-access
            )

    def test_fetch_product_info(self):
        """Test `_fetch_product_info` static method"""
        # Product name and version available.
        with patch('archey.entries.model.open', mock_open(), create=True) as mock:
            mock.return_value.read.side_effect = [
                'MY-LAPTOP-NAME\n',
                'MY-LAPTOP-VERSION\n'
            ]

            self.assertEqual(
                Model._fetch_product_info(),  # pylint: disable=protected-access
                'MY-LAPTOP-NAME MY-LAPTOP-VERSION'
            )

        # Only product name is available.
        with patch('archey.entries.model.open', mock_open(), create=True) as mock:
            mock.return_value.read.side_effect = [
                'MY-LAPTOP-NAME\n',
                FileNotFoundError()
            ]

            self.assertEqual(
                Model._fetch_product_info(),  # pylint: disable=protected-access
                'MY-LAPTOP-NAME'
            )

        # Product name is available but product version is empty.
        with patch('archey.entries.model.open', mock_open(), create=True) as mock:
            mock.return_value.read.side_effect = [
                'MY-LAPTOP-NAME\n',
                '\n'
            ]

            self.assertEqual(
                Model._fetch_product_info(),  # pylint: disable=protected-access
                'MY-LAPTOP-NAME'
            )

        # Neither product name nor version are available.
        with patch('archey.entries.model.open', mock_open(), create=True) as mock:
            mock.return_value.read.side_effect = [
                FileNotFoundError(),
                FileNotFoundError()
            ]

            self.assertIsNone(Model._fetch_product_info())  # pylint: disable=protected-access

    def test_fetch_rasperry_pi_revision(self):
        """Test `_fetch_rasperry_pi_revision` static method"""
        with patch('archey.entries.model.open', mock_open(), create=True) as mock:
            mock.return_value.read.side_effect = [
                'Hardware\t: HARDWARE\nRevision\t: REVISION\n',
                'processor   : 0\ncpu family  : X\n'
            ]

            self.assertEqual(
                Model._fetch_rasperry_pi_revision(),  # pylint: disable=protected-access
                'Raspberry Pi HARDWARE (Rev. REVISION)'
            )
            self.assertIsNone(
                Model._fetch_rasperry_pi_revision()  # pylint: disable=protected-access
            )

    @patch(
        'archey.entries.model.check_output',
        side_effect=[
            'PHONE-BRAND\n',     # First `getprop` call.
            'PHONE-DEVICE\n',    # Second `getprop` call.
            FileNotFoundError()  # Second test will fail.
        ]
    )
    def test_fetch_android_device_model(self, _):
        """Test `_fetch_android_device_model` static method"""
        self.assertEqual(
            Model._fetch_android_device_model(),  # pylint: disable=protected-access
            'PHONE-BRAND (PHONE-DEVICE)'
        )
        self.assertIsNone(
            Model._fetch_android_device_model()  # pylint: disable=protected-access
        )

    @patch(
        'archey.entries.model.Model._fetch_virtual_env_info',
        return_value=None  # Not a virtual environment...
    )
    @patch(
        'archey.entries.model.Model._fetch_product_info',
        return_value=None  # No model name could be retrieved...
    )
    @patch(
        'archey.entries.model.Model._fetch_rasperry_pi_revision',
        return_value=None  # Not a Raspberry Pi device either...
    )
    @patch(
        'archey.entries.model.Model._fetch_android_device_model',
        return_value=None  # Not an Android device...
    )
    @HelperMethods.patch_clean_configuration
    def test_no_match(self, _, __, ___, ____):
        """Test when no information could be retrieved"""
        model = Model()

        output_mock = MagicMock()
        model.output(output_mock)

        self.assertIsNone(model.value)
        self.assertEqual(
            output_mock.append.call_args[0][1],
            DEFAULT_CONFIG['default_strings']['not_detected']
        )


if __name__ == '__main__':
    unittest.main()



"""Archey API module"""

from datetime import datetime

import json

from archey._version import __version__


class API:
    """
    This class provides results serialization for external usages.
    At the moment, only JSON has been implemented.
    Feel free to contribute to add other formats as needed.
    """
    def __init__(self, entries):
        self.entries = entries

    def json_serialization(self, indent=None):
        """
        JSON serialization of entries.
        Set `indent` to the number of wanted output indentation tabs (2-space long).

        Note: For Python < 3.6, the keys order is not guaranteed.
        """
        document = {
            'data': {},
            'meta': {
                'version': self._version_to_semver_segments(__version__),
                'date': datetime.now().isoformat(),
                'count': len(self.entries)
            }
        }
        for entry in self.entries:
            document['data'][entry.name] = entry.value

        return json.dumps(
            document,
            indent=((indent * 2) or None)
        )

    @staticmethod
    def _version_to_semver_segments(version):
        """Transforms string `version` to a tuple containing SemVer segments"""
        return tuple(
            map(
                int,
                version.lstrip('v').partition('-')[0].split('.')
            )
        )



"""Processes entry class"""

from archey.entry import Entry
from archey.processes import Processes as ProcessesUtil


class Processes(Entry):
    """
    Simple wrapper to `archey.processes` to provide the number of running processes as an entry.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.value = ProcessesUtil().number



"""Test module for Archey's GPU detection module"""

import unittest
from unittest.mock import MagicMock, patch

from subprocess import CalledProcessError

from archey.entries.gpu import GPU
from archey.test import CustomAssertions
from archey.test.entries import HelperMethods
from archey.constants import DEFAULT_CONFIG


class TestGPUEntry(unittest.TestCase, CustomAssertions):
    """Here, we mock the `check_output` call to `lspci` to test the logic"""
    @patch(
        'archey.entries.gpu.check_output',
        return_value="""\
XX:YY.H IDE interface: IIIIIIIIIIIIIIII
XX:YY.H SMBus: BBBBBBBBBBBBBBBB
XX:YY.H VGA compatible controller: GPU-MODEL-NAME
XX:YY.H Audio device: DDDDDDDDDDDDDDDD
""")
    @HelperMethods.patch_clean_configuration
    def test_match_vga(self, _):
        """Simple 'VGA' device type matching"""
        self.assertListEqual(GPU().value, ['GPU-MODEL-NAME'])

    @patch(
        'archey.entries.gpu.check_output',
        return_value="""\
XX:YY.H IDE interface: IIIIIIIIIIIIIIII
XX:YY.H SMBus: BBBBBBBBBBBBBBBB
XX:YY.H VGA compatible controller: GPU-MODEL-NAME
XX:YY.H Display controller: ANOTHER-MATCHING-VIDEO-CONTROLLER
XX:YY.H Audio device: DDDDDDDDDDDDDDDD
XX:YY.H 3D controller: 3D GPU-MODEL-NAME TAKES ADVANTAGE
""")
    @HelperMethods.patch_clean_configuration(
        configuration={
            'gpu': {
                'one_line': True,
                'max_count': 2
            }
        }
    )
    def test_multi_matches_capped_one_line(self, _):
        """
        Test detection when there are multiple graphical device candidates.
        Check that `max_count` and `one_line` are honored too, including on `output` overriding.
        """
        gpu = GPU()

        output_mock = MagicMock()
        gpu.output(output_mock)

        self.assertListEqual(
            gpu.value,
            ['3D GPU-MODEL-NAME TAKES ADVANTAGE', 'GPU-MODEL-NAME']
        )
        self.assertEqual(
            output_mock.append.call_args[0][1],
            '3D GPU-MODEL-NAME TAKES ADVANTAGE, GPU-MODEL-NAME'
        )

    @patch(
        'archey.entries.gpu.check_output',
        return_value="""\
XX:YY.H IDE interface: IIIIIIIIIIIIIIII
XX:YY.H SMBus: BBBBBBBBBBBBBBBB
XX:YY.H Display controller: ANOTHER-MATCHING-VIDEO-CONTROLLER
XX:YY.H Audio device: DDDDDDDDDDDDDDDD
XX:YY.H 3D controller: 3D GPU-MODEL-NAME TAKES ADVANTAGE
""")
    @HelperMethods.patch_clean_configuration(
        configuration={
            'gpu': {
                'one_line': False,
                'max_count': False
            }
        }
    )
    def test_multi_matches_uncapped_multiple_lines(self, _):
        """
        Test detection when there are multiple graphical device candidates.
        Check that `max_count` and `one_line` are honored too, including on `output` overriding.
        """
        gpu = GPU()

        output_mock = MagicMock()
        gpu.output(output_mock)

        self.assertListEqual(
            gpu.value,
            ['3D GPU-MODEL-NAME TAKES ADVANTAGE', 'ANOTHER-MATCHING-VIDEO-CONTROLLER']
        )
        self.assertEqual(
            output_mock.append.call_args_list[0][0][1],
            '3D GPU-MODEL-NAME TAKES ADVANTAGE'
        )
        self.assertEqual(
            output_mock.append.call_args_list[1][0][1],
            'ANOTHER-MATCHING-VIDEO-CONTROLLER'
        )

    @patch(
        'archey.entries.gpu.check_output',
        return_value="""\
XX:YY.H IDE interface: IIIIIIIIIIIIIIII
XX:YY.H SMBus: BBBBBBBBBBBBBBBB
XX:YY.H Audio device: DDDDDDDDDDDDDDDD
""")
    @HelperMethods.patch_clean_configuration
    def test_no_match(self, _):
        """Test (non-)detection when there is not any graphical candidate"""
        gpu = GPU()

        output_mock = MagicMock()
        gpu.output(output_mock)

        self.assertListEmpty(gpu.value)
        self.assertEqual(
            output_mock.append.call_args[0][1],
            DEFAULT_CONFIG['default_strings']['not_detected']
        )

    @patch(
        'archey.entries.gpu.check_output',
        side_effect=CalledProcessError(1, 'lspci')
    )
    @HelperMethods.patch_clean_configuration
    def test_lspci_crash(self, _):
        """Test (non-)detection due to a crashing `lspci` program"""
        gpu = GPU()

        output_mock = MagicMock()
        gpu.output(output_mock)

        self.assertListEmpty(gpu.value)
        self.assertEqual(
            output_mock.append.call_args[0][1],
            DEFAULT_CONFIG['default_strings']['not_detected']
        )


if __name__ == '__main__':
    unittest.main()



"""Test module for Archey's kernel information detection module"""

import unittest
from unittest.mock import patch

from archey.entries.kernel import Kernel


class TestKernelEntry(unittest.TestCase):
    """
    Here, we mock the `check_output` call and check afterwards
      that the output is correct.
    """
    @patch(
        'archey.entries.kernel.check_output',
        return_value="""\
X.Y.Z-R-arch
""")
    def test(self, _):
        """A simple test, for a simple mock"""
        self.assertEqual(Kernel().value, 'X.Y.Z-R-arch')


if __name__ == '__main__':
    unittest.main()



"""Android logo"""

ANDROID = [
    """{c[0]}          -o          o-            """,
    """{c[0]}           +hydNNNNdyh+             """,
    """{c[0]}         +mMMMMMMMMMMMMm+           """,
    """{c[0]}       `dMM{c[1]}m:{c[0]}NMMMMMMN{c[1]}:m{c[0]}MMd`         """,
    """{c[0]}       hMMMMMMMMMMMMMMMMMMh         """,
    """{c[0]}   ..  yyyyyyyyyyyyyyyyyyyy  ..     """,
    """{c[0]} .mMMm`MMMMMMMMMMMMMMMMMMMM`mMMm.   """,
    """{c[0]} :MMMM-MMMMMMMMMMMMMMMMMMMM-MMMM:   """,
    """{c[0]} :MMMM-MMMMMMMMMMMMMMMMMMMM-MMMM:   """,
    """{c[0]} :MMMM-MMMMMMMMMMMMMMMMMMMM-MMMM:   """,
    """{c[0]} :MMMM-MMMMMMMMMMMMMMMMMMMM-MMMM:   """,
    """{c[0]} -MMMM-MMMMMMMMMMMMMMMMMMMM-MMMM-   """,
    """{c[0]}  +yy+ MMMMMMMMMMMMMMMMMMMM +yy+    """,
    """{c[0]}       mMMMMMMMMMMMMMMMMMMm         """,
    """{c[0]}       `/++MMMMh++hMMMM++/`         """,
    """{c[0]}           MMMMo  oMMMM             """,
    """{c[0]}           MMMMo  oMMMM             """,
    """{c[0]}           oNMm-  -mMNs             """
]



"""Simple `__init__` file for the `logos` module, to load each distribution logo"""

from archey.logos.alpine_linux import ALPINE_LINUX
from archey.logos.android import ANDROID
from archey.logos.arch_linux import ARCH_LINUX
from archey.logos.bunsenlabs import BUNSENLABS
from archey.logos.centos import CENTOS
from archey.logos.crunchbang import CRUNCHBANG
from archey.logos.debian import DEBIAN
from archey.logos.elementary import ELEMENTARY
from archey.logos.fedora import FEDORA
from archey.logos.freebsd import FREEBSD
from archey.logos.gentoo import GENTOO
from archey.logos.kali_linux import KALI_LINUX
from archey.logos.manjaro import MANJARO
from archey.logos.nixos import NIXOS
from archey.logos.linux import LINUX
from archey.logos.linux_mint import LINUX_MINT
from archey.logos.pop import POP
from archey.logos.openbsd import OPENBSD
from archey.logos.opensuse import OPENSUSE
from archey.logos.raspbian import RASPBIAN
from archey.logos.red_hat import RED_HAT
from archey.logos.slackware import SLACKWARE
from archey.logos.ubuntu import UBUNTU
from archey.logos.windows import WINDOWS


def get_logo_width(logo, nb_colors=8):
    """
    Utility function computing the real width of a distribution logo.
    Rationale : We use placeholders to dynamically set ANSI colors.
                Although, they **DO NOT** take width space once the text as been printed.

    For performance purposes we compute the logo length based on its first line.
    See `archey.test.test_archey_logos` unit tests for further logos consistency verifications.

    `logo` is supposed to be one of the constants declared above.
    `nb_colors` must be greater than or equal to the number of colors used by the logo.
    """
    # We replace each placeholder by a 0-character string.
    return len(logo[0].format(c=[''] * nb_colors))



"""Test module for Archey's device host-name detection module"""

import unittest
from unittest.mock import mock_open, patch

from archey.entries.hostname import Hostname


class TestHostnameEntry(unittest.TestCase):
    """Test cases mocking `/etc/hostname` or `check_output` call to `hostname`"""
    @patch(
        'archey.entries.hostname.open',
        mock_open(
            read_data="""\
MY-COOL-LAPTOP
"""),
        create=True
    )
    def test_etc_hostname(self):
        """Mock reading from `/etc/hostname`"""
        self.assertEqual(Hostname().value, 'MY-COOL-LAPTOP')

    @patch(
        'archey.entries.hostname.open',
        side_effect=FileNotFoundError(),
        create=True
    )
    @patch(
        'archey.entries.hostname.check_output',
        return_value="""\
MY-COOL-LAPTOP
""")
    def test_hostname(self, _, __):
        """Mock call to `hostname`"""
        self.assertEqual(Hostname().value, 'MY-COOL-LAPTOP')


if __name__ == '__main__':
    unittest.main()



"""Simple class (acting as a singleton) to handle processes listing"""

import sys

from subprocess import check_output

from archey.singleton import Singleton


class Processes(metaclass=Singleton):
    """At startup, instantiate this class to populate a list of running processes"""
    def __init__(self):
        try:
            self._processes = check_output(
                ['ps', '-eo', 'comm'],
                universal_newlines=True
            ).splitlines()[1:]
        except FileNotFoundError:
            sys.exit("Please, install first `procps` (or `procps-ng`) on your system.")

    @property
    def list(self):
        """Simple getter to retrieve (am immutable copy of) the processes list"""
        return tuple(self._processes)

    @property
    def number(self):
        """Simple getter to retrieve the number of stored processes"""
        return len(self._processes)



"""RAM usage detection class"""

import re

from subprocess import check_output

from archey.colors import Colors
from archey.entry import Entry


class RAM(Entry):
    """
    First tries to use the `free` command to retrieve RAM usage.
    If not available, falls back on the parsing of `/proc/meminfo` file.
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        used, total = self._get_used_total_values()
        if not total:
            return

        self.value = {
            'used': used,
            'total': total,
            'unit': 'MiB'
        }

    def _get_used_total_values(self):
        """
        Returns a tuple containing used and total RAM values.
        Tries a variety of methods, increasing compatibility for a wide range of systems.
        """
        try:
            return self._run_free_dash_m()
        except (IndexError, FileNotFoundError):
            pass

        try:
            return self._read_proc_meminfo()
        except (PermissionError, FileNotFoundError):
            pass

        return None, None

    @staticmethod
    def _run_free_dash_m():
        """Call `free -m` and parse its output to retrieve current used and total RAM"""
        memory_usage = ''.join(
            filter(
                re.compile('Mem').search,
                check_output(
                    ['free', '-m'],
                    env={'LANG': 'C'}, universal_newlines=True
                ).splitlines()
            )
        ).split()

        return float(memory_usage[2]), float(memory_usage[1])

    @staticmethod
    def _read_proc_meminfo():
        """Same behavior but by reading from `/proc/meminfo` directly"""
        with open('/proc/meminfo') as f_mem_info:
            mem_info_lines = f_mem_info.read().splitlines()

        # Store memory information into a dictionary.
        mem_info = {}
        for line in filter(None, mem_info_lines):
            key, value = line.split(':', maxsplit=1)
            mem_info[key] = float(value.strip(' kB')) / 1024

        total = mem_info['MemTotal']
        # Here, let's imitate Neofetch's behavior.
        # See <https://github.com/dylanaraps/neofetch/wiki/Frequently-Asked-Questions>.
        used = total + mem_info['Shmem'] - (
            mem_info['MemFree'] + mem_info['Cached']
            + mem_info['SReclaimable'] + mem_info['Buffers']
        )
        # Imitates what `free` does when the obtained value happens to be incorrect.
        # See <https://gitlab.com/procps-ng/procps/blob/master/proc/sysinfo.c#L790>.
        if used < 0:
            used = total - mem_info['MemFree']

        return used, total


    def output(self, output):
        """
        Adds the entry to `output` after pretty-formatting the RAM usage with color and units.
        """
        if not self.value:
            # Fall back on the default behavior if no RAM usage could be detected.
            super().output(output)
            return

        # DRY some constants
        used = self.value['used']
        total = self.value['total']
        unit = self.value['unit']
        # Fetch the user-defined RAM limits from configuration.
        ram_limits = self._configuration.get('limits')['ram']

        # Based on the RAM percentage usage, select the corresponding level color.
        level_color = Colors.get_level_color(
            (used / total) * 100,
            ram_limits['warning'], ram_limits['danger']
        )

        output.append(
            self.name,
            '{0}{1} {unit}{2} / {3} {unit}'.format(
                level_color,
                int(used),
                Colors.CLEAR,
                int(total),
                unit=unit
            )
        )



"""Test module for Archey's desktop environment detection module"""

import unittest
from unittest.mock import patch

from archey.entries.desktop_environment import DesktopEnvironment


class TestDesktopEnvironmentEntry(unittest.TestCase):
    """
    With the help of a fake running processes list, we test the DE matching.
    """
    @patch(
        'archey.entries.desktop_environment.Processes.list',
        (  # Fake running processes list
            'do',
            'you',
            'like',
            'cinnamon',  # Match !
            'tea'
        )
    )
    def test_match(self):
        """Simple list matching"""
        self.assertEqual(DesktopEnvironment().value, 'Cinnamon')

    @patch(
        'archey.entries.desktop_environment.Processes.list',
        (  # Fake running processes list
            'do',
            'you',
            'like',
            'unsweetened',  # Mismatch...
            'coffee'
        )
    )
    @patch(
        'archey.entries.desktop_environment.os.getenv',
        return_value='DESKTOP ENVIRONMENT'
    )
    def test_mismatch(self, _):
        """Simple list (mis-)-matching"""
        self.assertEqual(DesktopEnvironment().value, 'DESKTOP ENVIRONMENT')

    @patch(
        'archey.entries.desktop_environment.Processes.list',
        (  # Fake running processes list
            'do',
            'you',
            'like',
            'unsweetened',  # Mismatch...
            'coffee'
        )
    )
    @patch(
        'archey.entries.desktop_environment.os.getenv',
        return_value=None  # The environment variable is empty...
    )
    def test_non_detection(self, _):
        """Simple global non-detection"""
        self.assertIsNone(DesktopEnvironment().value)


if __name__ == '__main__':
    unittest.main()



"""
Distributions enumeration module.
Operating Systems detection logic.
Interface to `os-release` (through `distro` module).
"""

import os
import sys

from enum import Enum
from subprocess import check_output

import distro


class Distributions(Enum):
    """
    This enumeration lists supported operating systems (keys).
    Values contain their respective `distro` identifier.
    See <https://distro.readthedocs.io/en/latest/#distro.id>.
    """
    ALPINE_LINUX = 'alpine'
    ANDROID = 'android'
    ARCH_LINUX = 'arch'
    BUNSENLABS = 'bunsenlabs'
    CENTOS = 'centos'
    CRUNCHBANG = 'crunchbang'
    DEBIAN = 'debian'
    ELEMENTARY = 'elementary'
    FEDORA = 'fedora'
    FREEBSD = 'freebsd'
    GENTOO = 'gentoo'
    KALI_LINUX = 'kali'
    MANJARO_LINUX = 'manjaro'
    NIXOS = 'nixos'
    LINUX = 'linux'
    LINUX_MINT = 'linuxmint'
    OPENBSD = 'openbsd'
    OPENSUSE = 'opensuse'
    POP = 'pop'
    RASPBIAN = 'raspbian'
    RED_HAT = 'rhel'
    SLACKWARE = 'slackware'
    UBUNTU = 'ubuntu'
    WINDOWS = 'windows'


    @staticmethod
    def get_distribution_identifiers():
        """Simple getter returning current supported distributions identifiers"""
        return [d.value for d in Distributions.__members__.values()]

    @staticmethod
    def run_detection():
        """Entry point of Archey distribution detection logic"""
        distribution = Distributions._detection_logic()

        # In case nothing got detected the "regular" way, fall-back on the Linux logo.
        if not distribution:
            # Android systems are currently not being handled by `distro`.
            # We imitate Neofetch behavior to manually "detect" them.
            # See <https://github.com/nir0s/distro/issues/253>.
            if os.path.isdir('/system/app') and os.path.isdir('/system/priv-app'):
                return Distributions.ANDROID

            return Distributions.LINUX

        # Below are brain-dead cases for distributions not properly handled by `distro`.
        # One _may_ want to add its own logic to add support for such undetectable systems.
        if distribution == Distributions.DEBIAN:
            # CrunchBang is tagged as _regular_ Debian by `distro`.
            # Below conditions are here to work-around this issue.
            # First condition : CrunchBang-Linux and CrunchBang-Monara.
            # Second condition : CrunchBang++ (CBPP).
            if os.path.isfile('/etc/lsb-release-crunchbang') \
                or os.path.isfile('/usr/bin/cbpp-exit'):
                return Distributions.CRUNCHBANG

        elif distribution == Distributions.UBUNTU:
            # Older Pop!_OS releases (< 20.*) didn't ship their own `ID` (from `os-release`).
            # Thus, they are detected as "regular" Ubuntu distributions.
            # We may here rely on their `NAME` (from `os-release`), which is sufficient.
            if Distributions.get_distro_name(pretty=False) == 'Pop!_OS':
                return Distributions.POP

        return distribution

    @staticmethod
    def _detection_logic():
        """Main distribution detection logic, relying on `distro`, handling _common_ cases"""
        # Are we running on Windows ?
        if sys.platform in ('win32', 'cygwin'):
            return Distributions.WINDOWS

        # Is it a Windows Sub-system Linux (WSL) distribution ?
        # If so, kernel release identifier should keep a trace of it.
        if b'microsoft' in check_output(['uname', '-r']).lower():
            return Distributions.WINDOWS

        # Is `ID` (from `os-release`) well-known and supported ?
        try:
            return Distributions(distro.id())
        except ValueError:
            pass

        # Is any of `ID_LIKE` (from `os-release`) well-known and supported ?
        # See <https://www.freedesktop.org/software/systemd/man/os-release.html#ID_LIKE=>.
        for id_like in distro.like().split(' '):
            try:
                return Distributions(id_like)
            except ValueError:
                pass

        # Nothing of the above matched, let's return `None` and let the caller handle it.
        return None

    @staticmethod
    def get_distro_name(pretty=True):
        """Simple wrapper to `distro` to return the current distribution _pretty_ name"""
        return distro.name(pretty=pretty) or None

    @staticmethod
    def get_ansi_color():
        """
        Simple wrapper to `distro` to return the distribution preferred ANSI color.
        See <https://www.freedesktop.org/software/systemd/man/os-release.html#ANSI_COLOR=>.
        """
        return distro.os_release_attr('ansi_color') or None



"""Test module for Archey's session user name detection module"""

from subprocess import CalledProcessError

import unittest
from unittest.mock import MagicMock, patch

from archey.entries.user import User
from archey.test.entries import HelperMethods
from archey.constants import DEFAULT_CONFIG


class TestUserEntry(unittest.TestCase):
    """
    For this entry, we'll check the output by mocking the `getenv` call.
    If the environment variable happens to be empty, we fall-back to a fake `id` call.
    """
    @patch(
        'archey.entries.user.os.getenv',
        return_value='USERNAME'
    )
    def test_getenv(self, _):
        """Simple mock, simple test"""
        self.assertEqual(User().value, 'USERNAME')

    @patch(
        'archey.entries.user.os.getenv',
        return_value=None
    )
    @patch(
        'archey.entries.user.check_output',
        return_value='USERNAME\n'
    )
    def test_id_call(self, _, __):
        """Mock `id` returned value and check the correct assignment"""
        self.assertEqual(User().value, 'USERNAME')

    @patch(
        'archey.entries.user.os.getenv',
        return_value=None
    )
    @patch(
        'archey.entries.user.check_output',
        side_effect=CalledProcessError(1, 'id', "id: ’1000’: no such user\n")
    )
    @HelperMethods.patch_clean_configuration
    def test_config_fall_back(self, _, __):
        """`id` fails, but Archey must not !"""
        user = User()

        output_mock = MagicMock()
        user.output(output_mock)

        self.assertIsNone(user.value)
        self.assertEqual(
            output_mock.append.call_args[0][1],
            DEFAULT_CONFIG['default_strings']['not_detected']
        )


if __name__ == '__main__':
    unittest.main()



"""CPU information detection class"""

import re

from subprocess import check_output

from archey.entry import Entry


class CPU(Entry):
    """
    Parse `/proc/cpuinfo` file to retrieve the model name.
    If no information could be retrieved, calls `lscpu`.
    """
    _MODEL_NAME_REGEXP = re.compile(
        r'^model name\s*:\s*(.*)$',
        flags=re.IGNORECASE | re.MULTILINE
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        cpuinfo_match = self._read_proc_cpuinfo()
        if not cpuinfo_match:
            # This test case has been built for some ARM architectures (see #29).
            # Sometimes, `model name` info is not present within `/proc/cpuinfo`.
            # We use the output of `lscpu` program (util-linux-ng) to retrieve it.
            cpuinfo_match = self._run_lscpu()

        # Sometimes CPU model name contains extra ugly white-spaces.
        self.value = re.sub(r'\s+', ' ', cpuinfo_match.group(1))

    def _read_proc_cpuinfo(self):
        """Read `/proc/cpuinfo` and search for our model name pattern"""
        try:
            with open('/proc/cpuinfo') as f_cpu_info:
                return self._MODEL_NAME_REGEXP.search(f_cpu_info.read())
        except (PermissionError, FileNotFoundError):
            return None

    def _run_lscpu(self):
        """Same operation but from `lscpu` output"""
        return self._MODEL_NAME_REGEXP.search(
            check_output(
                ['lscpu'],
                env={'LANG': 'C'}, universal_newlines=True
            )
        )



#!/usr/bin/env python3

"""
This is the Archey 4's `setup.py` file, allowing us to distribute it as a package...
... with cool meta-data.
"""

import os

from setuptools import find_packages, setup

from archey._version import __version__


setup(
    name='archey4',
    version=__version__.lstrip('v'),
    description='Archey is a simple system information tool written in Python',
    keywords='archey python3 linux system-information monitoring screenshot',
    url='https://github.com/HorlogeSkynet/archey4',
    author='Samuel Forestier',  # Not alone
    author_email='dev+archey@samuel.domains',
    license='GPLv3',
    packages=find_packages(exclude=['archey.test']),
    test_suite='archey.test',
    python_requires='>=3.4',
    install_requires=[
        'distro',
        'netifaces'
    ],
    entry_points={
        'console_scripts': [
            'archey = archey.__main__:main'
        ]
    },
    long_description="""\
Archey4 is a **maintained** fork of the original Archey Linux system tool.
The original Archey program had been written by Melik Manukyan in 2009, and quickly abandoned in 2011.
At first, it only supported Arch Linux distribution, further support had been added afterwards.
Many forks popped in the wild due to inactivity, but this one attends since 2017 to succeed where the others failed:
Remain *maintained*, *community-driven* and *highly-compatible* with yesterday's and today's systems.\
""",
    long_description_content_type='text/x-rst',
    data_files=[
        # By filtering on `os.path.exists`, install should succeed even when
        #   the compressed manual page is not available (iterable would be empty).
        ('share/man/man1', filter(os.path.exists, ['dist/archey.1.gz'])),
        ('share/doc/archey4', ['README.md', 'COPYRIGHT.md'])
    ],
    zip_safe=False,
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Natural Language :: English',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Topic :: System'
    ]
)



"""Host name detection class"""

from subprocess import check_output

from archey.entry import Entry


class Hostname(Entry):
    """Simple call to `uname` to retrieve the host name"""
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.value = self._read_etc_hostname()
        if not self.value:
            self.value = check_output(
                ['uname', '-n'],
                universal_newlines=True
            ).rstrip()

    @staticmethod
    def _read_etc_hostname():
        try:
            with open('/etc/hostname') as f_hostname:
                return f_hostname.read().rstrip()
        except FileNotFoundError:
            return None



"""Test module for Archey's terminal detection module"""

import unittest
from unittest.mock import MagicMock, patch

from archey.entries.terminal import Terminal
from archey.test.entries import HelperMethods
from archey.constants import DEFAULT_CONFIG


@patch(
    'archey.entries.terminal.NO_COLOR',
    False  # By default, colors won't be disabled.
)
class TestTerminalEntry(unittest.TestCase):
    """
    For this entry, we'll verify that the output contains what the environment
      is supposed to give, plus the right number of "colorized" characters.
    """
    @patch.dict(
        'archey.entries.terminal.os.environ',
        {
            'TERM': 'xterm-256color',
            'COLORTERM': 'truecolor',
            'TERM_PROGRAM': 'A-COOL-TERMINAL-EMULATOR'
        },
        clear=True
    )
    @HelperMethods.patch_clean_configuration
    def test_terminal_emulator_term_program(self):
        """Check that `TERM_PROGRAM` is honored even if `TERM` or `COLORTERM` is defined"""
        self.assertEqual(Terminal().value, 'A-COOL-TERMINAL-EMULATOR')

    @patch.dict(
        'archey.entries.terminal.os.environ',
        {
            'TERM': 'OH-A-SPECIAL-CASE',
            'TERMINATOR_UUID': 'urn:uuid:xxxxxxxx-yyyy-zzzz-tttt-uuuuuuuuuuuu'  # Ignored.
        },
        clear=True
    )
    @HelperMethods.patch_clean_configuration
    def test_terminal_emulator_special_term(self):
        """Check that `TERM` is honored even if a "known identifier" could be found"""
        self.assertEqual(Terminal().value, 'OH-A-SPECIAL-CASE')

    @patch.dict(
        'archey.entries.terminal.os.environ',
        {
            'TERM': 'xterm-256color',
            'KONSOLE_VERSION': 'X.Y.Z'
        },
        clear=True
    )
    @HelperMethods.patch_clean_configuration
    def test_terminal_emulator_name_normalization(self):
        """Check that our manual terminal detection as long as name normalization are working"""
        self.assertEqual(Terminal().value, 'Konsole')

    @patch.dict(
        'archey.entries.terminal.os.environ',
        {'TERM': 'xterm-256color'},
        clear=True
    )
    @HelperMethods.patch_clean_configuration(
        configuration={
            'colors_palette': {
                'use_unicode': True
            }
        }
    )
    def test_terminal_emulator_term_fallback_and_unicode(self):
        """Check that `TERM` is honored if present, and Unicode support for the colors palette"""
        terminal = Terminal()

        output_mock = MagicMock()
        terminal.output(output_mock)

        self.assertEqual(terminal.value, 'xterm-256color')
        self.assertTrue(
            output_mock.append.call_args[0][1].startswith('xterm-256color')
        )
        self.assertEqual(
            output_mock.append.call_args[0][1].count('\u2588'),
            7 * 2
        )

    @patch.dict(
        'archey.entries.terminal.os.environ',
        {'COLORTERM': 'kmscon'},
        clear=True
    )
    @HelperMethods.patch_clean_configuration
    def test_terminal_emulator_colorterm(self):
        """Check we can detect terminals using the `COLORTERM` environment variable."""
        self.assertEqual(Terminal().value, 'KMSCON')

    @patch.dict(
        'archey.entries.terminal.os.environ',
        {
            'TERM': 'xterm-256color',
            'KONSOLE_VERSION': '200401',
            'COLORTERM': 'kmscon'
        },
        clear=True
    )
    @HelperMethods.patch_clean_configuration
    def test_terminal_emulator_colorterm_override(self):
        """
        Check we observe terminal using `COLORTERM` even if `TERM` or a "known identifier" is found.
        """
        self.assertEqual(Terminal().value, 'KMSCON')

    @patch.dict(
        'archey.entries.terminal.os.environ',
        {'TERM_PROGRAM': 'X-TERMINAL-EMULATOR'},
        clear=True
    )
    def test_no_color(self):
        """Test `Terminal` output behavior when `NO_COLOR` is set (palette should be hidden)"""
        with patch(
                'archey.entries.terminal.NO_COLOR',
                True  # Temporary disable color for this test.
            ):
            terminal = Terminal()

            output_mock = MagicMock()
            terminal.output(output_mock)

            self.assertEqual(terminal.value, 'X-TERMINAL-EMULATOR')
            self.assertEqual(
                output_mock.append.call_args[0][1],
                'X-TERMINAL-EMULATOR'
            )

    @patch.dict(
        'archey.entries.terminal.os.environ',
        {},
        clear=True
    )
    @HelperMethods.patch_clean_configuration(
        configuration={
            'colors_palette': {
                'use_unicode': False
            }
        }
    )
    def test_not_detected(self):
        """Test terminal emulator (non-)detection, without Unicode support"""
        terminal = Terminal()

        output_mock = MagicMock()
        terminal.output(output_mock)
        output = output_mock.append.call_args[0][1]

        self.assertIsNone(terminal.value)
        self.assertTrue(output.startswith(DEFAULT_CONFIG['default_strings']['not_detected']))
        self.assertFalse(output.count('\u2588'))


if __name__ == '__main__':
    unittest.main()
